#Michael Zakrewsky
#CHE210D - Final Project - Stretching DNA
#06/15/12

#import python modules
import numpy as np
import time
import atomwrite

#import compiled fortran library
import dnalib


def InitPositions(N, L, Lc):
    """Returns an array of initial positions of each atom
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    #compute integer grid # of locations for cubic lattice
    NLat = int(L * (N - 1)/Lc)
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float)
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0            
    for y in r:
        Pos[i] = np.array([L/2.,y + L/4.,L/2.], float)
        #add a random offset to help initial minimization
        Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
        Pos[i] = Pos[i] + Offset
        i += 1
        if i >= N:
            return Pos
    return Pos

def ThermoStat(Vel, T, N):
    """Introduces random velocity changes from a Boltzmann distribution
    to maintain the system at the appropriate temperature
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
    N: Number of particles
Output:
    Vel: same as above
"""
    #mass per particle
    m = (1.047e-19)/N
    sigma = np.sqrt(T * 1000000/m)
    Vel = np.random.normal(0,sigma,size=(N,3))
    Vel = Vel - Vel.mean(axis=0)
    return Vel

def RescaleVelocities(Vel, T, N):
    """Rescales velocities in the system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
    N: Number of particles
Output:
    Vel: same as above
"""
    #mass per particle
    m = (1.047e-19)/N    
    #recenter to zero net momentum (assuming all masses same)
    Vel = Vel - Vel.mean(axis=0)
    #find the total kinetic energy
    KE = 0.5 * m * np.sum(Vel * Vel)
    #find velocity scale factor from ratios of kinetic energy
    VScale = np.sqrt(1.5 * N * T * 1000000/KE)
    Vel = Vel * VScale
    return Vel


def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = RescaleVelocities(Vel, T, N)
    return Vel


def InitAccel(Pos):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #global variables
    global T, k, Gamma, Alpha, Ksi, Lambda, Lc, N
    #get the acceleration from the forces
    Forces = np.zeros_like(Pos)
    PEnergy, Forces = dnalib.calcenergyforces(Pos, T, Kappa1, Kappa2, Ext, Gamma, Alpha, Ksi, Lambda, Lc, Forces, 3, N)
    Accel = Forces * 1000000/((1.047e-19)/N)
    return Accel


#### VARIABLES FOR THE CODE ABOVE TO WORK ####
dt = 1.e-9                                       #time step in seconds
RescaleFreq = 250                                #velocity rescaling frequency
EquilSteps1 = 1500000                            #number of equilibration steps
EquilSteps2 = 2*EquilSteps1/3.
ProdSteps = EquilSteps2/2.                       #number of production steps               

N = 25                                           #total number of atoms
M = .010                                         #salt concentration (Molarity)
T = 4.278e-15                                    #kb * T at 37 degrees C (N*um)                 
Lc = 32.8                                        #DNA contour length (um)
L = 2 * Lc                                       #dimension of box (um)
Gamma = 365853                                   #stretching constant (um-2)
Alpha = 0.0388                                   #bending constant
Ksi = 46533                                      #electrostatic constant (um) 
Lambda = (1./np.sqrt(10.8 * M))/1000.            #debye length (um)
Kappa1 = 50000.                                  #tether constant (um-2)
Kappa2 = 10.                                     #umbrella sampling constant (um-2)
Ext = np.array([32.8,43.4,32.8],dtype=float)

#whether or not to use the 3d visualization
UseVisual = False
SaveTrajectory = True

if UseVisual:
    #import custom visualization library
    import atomvis

StartTime = time.time()

#print system parameters
print "Molarity = %.5f" % (M)
print "Debye-Length = %.3e" % (Lambda)

#z = np.zeros((ProdSteps/RescaleFreq,3), dtype=float)
m = 0
if SaveTrajectory:
    Pdb = atomwrite.pdbfile("animation10.pdb")
    
while m < 10:
    print Ext
    #print "Run = %d" % (m+1)
    #set initial conditions
    Pos = InitPositions(N, L, Lc)
    PosStart = Pos[0]
    Vel = InitVelocities(N, T)
    Accel = InitAccel(Pos)
    PEnergy, Forces = dnalib.calcenergyforces(Pos, T, Kappa1, Kappa2, Ext, Gamma, Alpha, Ksi, Lambda, Lc, Accel, 3, N)
    KEnergy = 0.5 * ((1.047e-19)/N) * np.sum(Vel * Vel)/1000000000000.
    print "Starting Contour Length = %.3f" % (np.sqrt(np.sum((Pos[0] - Pos[N-1])*(Pos[0] - Pos[N-1]))))
    print "Starting Potential Energy = %.3e" % (PEnergy)
    print "Starting Kinetic Energy = %.3e" % (KEnergy)

    #if using visualization, update the display
    if UseVisual:
        #initialize the visualization window
        atomvis.Init(Pos)

    #initialize and run MD equilibration1
    l = 0
    print "Start MD Equilibration1"

    while l < EquilSteps1:
        #do one integration step
        Pos, Vel, Accel, KEnergy, PEnergy = dnalib.vvintegrate(Pos, T, Kappa1, Kappa2, Ext, Vel, Accel, Gamma, Alpha, Ksi, Lambda, Lc, dt, 3, N)
        l+=1
        
        #check if we need to rescale the velocities
        if l % RescaleFreq == 0:
            Vel = ThermoStat(Vel, T, N)
        
            #update the 3D visualization
            if UseVisual:
                atomvis.Update(Pos, L)
        
    #initialize and run MD equilibration2
    l = 0
    print "Start MD Equilibration2"
        
    #print initial conditions
    #t = dt*(l)
    #print "%.3e,%.4e,%.4e,%.4e" % (t, PEnergy, KEnergy, PEnergy + KEnergy)
        
    while l < EquilSteps2:
        #do one integration step while keeping the dna tethered
        Pos, Vel, Accel, KEnergy, PEnergy = dnalib.vvintegrate(Pos, T, Kappa1, Kappa2, Ext, Vel, Accel, Gamma, Alpha, Ksi, Lambda, Lc, dt, 3, N)
        l+=1
    
        #check if we need to rescale the velocities
        if l % RescaleFreq == 0:
            Vel = ThermoStat(Vel, T, N)
        
            if UseVisual:
                atomvis.Update(Pos, L)
        
        #record PE, KE, and E every 10000 mdsteps
        #if l % 10000 == 0:
            #t = dt*(l)
            #print "%.3e,%.4e,%.4e,%.4e" % (t, PEnergy, KEnergy, PEnergy + KEnergy)

    #perform production run
    print "Start Production Run"

    l = 0
    q = 0

    while l < ProdSteps:
        #do one integration step while keeping the dna tethered
        Pos, Vel, Accel, KEnergy, PEnergy = dnalib.vvintegrate(Pos, T, Kappa1, Kappa2, Ext, Vel, Accel, Gamma, Alpha, Ksi, Lambda, Lc, dt, 3, N)
        l+=1
    
        #check if we need to rescale the velocities
        if l % RescaleFreq == 0:
            Vel = ThermoStat(Vel, T, N)
        
            if UseVisual:
                atomvis.Update(Pos, L)
        
        if l % RescaleFreq*10. == 0:
            if SaveTrajectory:
                Pdb.write(Pos)
                
            #z[q,m] = np.sqrt(np.sum((Pos[N-1] - Pos[0])*(Pos[N-1] - Pos[0])))
            #q+=1
        
        #record PE, KE, and E every 10000 mdsteps
        #if l % 10000 == 0:
            #t = dt*(l)
            #print "%.3e,%.4e,%.4e,%.4e" % (t, PEnergy, KEnergy, PEnergy + KEnergy)
    #zm = z[:,m]
    #print "%.5f,%.5f,%.5f,%.5f" % (zm.mean(),zm.std(),zm.min(),zm.max())      
    
    if UseVisual:
        atomvis.Close()
    
    if SaveTrajectory:
        Pdb.close()
    
    m+=1
    if m > 0:
        SaveTrajectory=False
#print "z distribution"
#i=0
#while i < len(z):
    #print "%.5f,%.5f,%.5f" % (z[i,0],z[i,1],z[i,2])
    #i+=1

EndTime = time.time()

print "Total Run Time (min) = %.3f" % ((EndTime - StartTime)/60.)